import { ClassType, Inject } from '@deepkit/core';
import { TransientInjectionTarget } from '@deepkit/injector';
import { MemoryLoggerTransport } from './memory-logger.js';
export declare enum LoggerLevel {
    none = 0,
    alert = 1,
    error = 2,
    warning = 3,
    log = 4,
    info = 5,
    debug = 6,
    debug2 = 7
}
export type LogData = {
    [name: string]: any;
};
export interface LogMessage {
    message: string;
    rawMessage: string;
    date: Date;
    level: LoggerLevel;
    scope: string;
    data: LogData;
}
export declare class ConsoleTransport implements LoggerTransport {
    protected withColors: boolean;
    constructor(withColors?: boolean);
    write(message: LogMessage): void;
    supportsColor(): boolean;
}
export declare class JSONTransport implements LoggerTransport {
    out: {
        write: (v: string) => any;
    };
    write(message: LogMessage): void;
    supportsColor(): boolean;
}
export interface LoggerTransport {
    write(message: LogMessage): void;
    supportsColor(): boolean;
}
export interface LoggerFormatter {
    format(message: LogMessage): void;
}
export declare class ColorFormatter implements LoggerFormatter {
    static colors: string[];
    format(message: LogMessage): void;
}
export declare class RemoveColorFormatter implements LoggerFormatter {
    format(message: LogMessage): void;
}
export declare class DefaultFormatter implements LoggerFormatter {
    formatters: LoggerFormatter[];
    format(message: LogMessage): void;
}
export declare class TimestampFormatter implements LoggerFormatter {
    format(message: LogMessage): void;
}
export declare class LogLevelFormatter implements LoggerFormatter {
    format(message: LogMessage): void;
}
export declare class ScopeFormatter implements LoggerFormatter {
    format(message: LogMessage): void;
}
export interface LoggerInterface {
    level: LoggerLevel;
    scoped(name: string): LoggerInterface;
    /**
     * Sends additional log data for the very next log/error/alert/warning/etc call.
     *
     * @example
     * ```typescript
     *
     * logger.data({user: user}).log('User logged in');
     *
     * //or
     *
     * //the given data is only used for the very next log (or error/alert/warning etc) call.
     * logger.data({user: user})
     * logger.log('User logged in');
     *
     * //at this point `data` is consumed, and for all other log calls not used anymore.
     * logger.log('another message without data');
     *
     *
     * ```
     */
    data(data: LogData): LoggerInterface;
    is(level: LoggerLevel): boolean;
    alert(...message: any[]): void;
    error(...message: any[]): void;
    warn(...message: any[]): void;
    log(...message: any[]): void;
    info(...message: any[]): void;
    debug(...message: any[]): void;
    debug2(...message: any[]): void;
}
export declare class Logger implements LoggerInterface {
    protected transporter: LoggerTransport[];
    protected formatter: LoggerFormatter[];
    scope: string;
    protected colorFormatter: ColorFormatter;
    protected removeColorFormatter: RemoveColorFormatter;
    /**
     * Setting a log level means only logs below or equal to this level will be handled.
     */
    level: LoggerLevel;
    protected scopeLevels: Map<string, LoggerLevel>;
    protected logData?: LogData;
    protected scopes: {
        [scope: string]: Logger;
    };
    constructor(transporter?: LoggerTransport[], formatter?: LoggerFormatter[], scope?: string);
    setLevel(level: keyof typeof LoggerLevel): void;
    isScopeEnabled(name: string): boolean;
    setScopeLevel(name: string, level: LoggerLevel): void;
    unsetScopeLevel(name: string): void;
    getScopeLevel(name: string): number;
    /**
     * Sends additional log data for the very next log/error/alert/warning/etc call.
     *
     * @example
     * ```typescript
     *
     * logger.data({user: user}).log('User logged in');
     *
     * //or
     *
     * //the given data is only used for the very next log (or error/alert/warning etc) call.
     * logger.data({user: user})
     * logger.log('User logged in');
     *
     * //at this point `data` is consumed, and for all other log calls not used anymore.
     * logger.log('another message without data');
     *
     *
     * ```
     */
    data(data: LogData): this;
    /**
     * Creates a new scoped logger. A scoped logger has the same log level, transports, and formatters as the parent logger,
     * and references them directly. This means if you change the log level on the parent logger, it will also change for all
     * scoped loggers.
     */
    scoped(name: string): Logger;
    addTransport(transport: LoggerTransport): void;
    setTransport(transport: LoggerTransport[]): void;
    removeTransport(transport: LoggerTransport): void;
    hasFormatter(formatterType: ClassType<LoggerFormatter>): boolean;
    hasFormatters(): boolean;
    addFormatter(formatter: LoggerFormatter): void;
    setFormatter(formatter: LoggerFormatter[]): void;
    protected format(message: LogMessage): void;
    is(level: LoggerLevel): boolean;
    protected send(messages: any[], level: LoggerLevel, data?: LogData): void;
    alert(...message: any[]): void;
    error(...message: any[]): void;
    warn(...message: any[]): void;
    log(...message: any[]): void;
    info(...message: any[]): void;
    debug(...message: any[]): void;
    debug2(...message: any[]): void;
}
/**
 * Logger with pre-configured console transport.
 */
export declare class ConsoleLogger extends Logger {
    constructor();
}
/**
 * Logger with pre-configured memory transport.
 */
export declare class MemoryLogger extends Logger {
    memory: MemoryLoggerTransport;
    constructor(transporter?: LoggerTransport[], formatter?: LoggerFormatter[], scope?: string);
    getOutput(): string;
    clear(): void;
}
export type ScopedLogger = Inject<Logger, 'scoped-logger'>;
export declare const ScopedLogger: {
    readonly provide: "scoped-logger";
    readonly transient: true;
    readonly useFactory: (target: TransientInjectionTarget, logger?: Logger) => Logger;
};
export declare type __ΩLoggerLevel = any[];
export declare type __ΩLogData = any[];
export declare type __ΩLogMessage = any[];
export declare type __ΩLoggerTransport = any[];
export declare type __ΩLoggerFormatter = any[];
export declare type __ΩLoggerInterface = any[];
export declare type __ΩScopedLogger = any[];
